########################################################################
#
# File Name:            Uri.py
#
"""
WWW: http://4suite.org/4Suite         e-mail: support@4suite.org

Copyright 2000-2002 Fourthought Inc., USA.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import urlparse, urllib, os, re
import Uuid
from Ft.Lib import UriException

#============================================================================
# Regular expressions for determining the non-URI-ness of strings
#

# Courtesy Altova (in the public domain)
# hacked from the BNF in RFC 2396 (http://www.ietf.org/rfc/rfc2396.txt)
# URI_REGEX = r"(([a-zA-Z][0-9a-zA-Z+\-\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?(#[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
#
# Altova's regex is actually very lenient and will match either a URI or
# a URI reference anywhere in a string. Since an empty string is a valid
# URI (it is zero or more URI-looking things), this results in *every*
# string matching the pattern! So, in order for a pattern that possibly
# matches an empty string to be useful, we have to be sure the pattern
# matches the entire string we are testing. So don't look for a URI ref
# *in* a string, look for a string that is, from start to finish, a URI
# ref, like this: ^URI_REGEX$
#
# Also, one must be careful to realize that a test like this only
# determines whether something matches the URI reference syntax, and
# is not a guarantee that it is actually a URI; it could just be an
# alphabetic string like "helloWorld" or "C:\autoexec.bat". All you can
# say with this test is that it definitely is not a URI, or that it
# "could be" a URI.
#
# As it might be more useful to test a string to see if it could be just
# a URI and not a URI reference, we'll derive from the Altova regex a
# pair of Python regular-expression objects, one for URIs and the other
# for URI references. We optimize it slightly by using "(?:foo)" instead
# of "(foo)" for grouping.
#

#Expressions moved to the init function


# Python regular expressions for strict validation of URI references
# by Mike Brown (Fourthought, Inc.)
#
# Public domain.
#
# Based on the BNF in RFC 2396, as amended by RFC 2732
#
# version: 2002-02-18T21:18:00
#
# The compiled patterns use the Python-specific feature of named groups, like
# "(?P<foo>bar)", so that the portion of the string matching the 'foo'
# group's pattern can be referenced later. This means one can use a Match
# object to see how the string was interpreted, and to access some of its
# components, if desired, although it may be easier to just reparse with
# splitUriRef(). These patterns also use discardable groups, indicated by
# "(?:foo)", for efficiency.
#
# Note that we do not use any \d or \w shortcuts, as these are
# potentially locale or Unicode sensitive.
#
# A non-Python version of the expressions could be made by doing a simple
# replacement, in the strings below
#    "(?:"  becomes  "("
#    "(?P<foo>"  where foo matches [a-zA-Z_]+  becomes  "("
# However, there are about 170 groups. This might overflow a regular
# expression parsing engine. Python only accepts 100.
#

#These patterns have been moved to the init function below



#============================================================================
# Functions to validate URIs and URI references
#
def matchesUriRefSyntax(s, strict=1):
    """
    This function returns 1 if a given string could be a URI reference (an
    absolute or relative URI, plus an optional #fragment) as defined in RFC
    2396, just based on the string's syntax.

    The 'strict' argument can be set to 0 for lenient, faster checking.

    Note that an empty string is a valid URI reference. Also note that a
    successful match does not guarantee that the string is a URI reference;
    only that it could be one, since it meets the syntax requirements.
    Failure to match indicates that it definitely is not a URI reference.
    """
    #FIXME, for some reason, "\n" is mathing as a valid URI ref.  This should not be the case
    #So, if the string consists of all white space, then return 0
    if s and not s.strip(): return 0

    _initUriRegex()
    if strict:
        return pyStrictUriRefPattern.match(s) is not None
    else:
        return pyUriRefPattern.match(s) is not None

def matchesUriSyntax(s, strict=1):
    """
    This function returns 1 if a given string could be an absolute or
    relative URI, as defined in RFC 2396, just based on the string's syntax.

    The 'strict' argument can be set to 0 for lenient, faster checking.

    Note that an empty string is a valid URI. Also note that a successful
    match does not guarantee that the string is a URI; only that it could
    be one, since it meets the syntax requirements. Failure to match
    indicates that it is definitely not a URI.
    """
    _initUriRegex()

    if strict:
        return pyStrictUriPattern.match(s) is not None
    else:
        return pyUriPattern.match(s) is not None

def isAbsolute(identifier):
    """
    Given a string believed to be a URI or URI reference, tests that the
    identifier is absolute (as per RFC 2396), not relative -- i.e., that
    it has a scheme.
    """
    # We do it this way to avoid compiling another massive regex.
    return splitUriRef(identifier)[0] is not None

def _normalizeStrictPatternGroups(matchGroupDict):
    """
    Given a dictionary as obtained with m.groupdict(), where m is the Match
    object returned by pyStrictUriRefPattern.match(s) or
    pyStrictUriPattern.match(s), returns a dictionary in which all keys
    with None values have been removed, and in which the _A or _R suffixes
    on the keys have been removed.

    (These suffixes exist because of the need to keep group names in the
    regular expressions unique, while still being able to use them multiple
    times within the expression).

    Example usage:

    from Ft.Lib import Uri
    s = 'http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right#toc'
    m = Uri.pyStrictUriRefPattern.match(s)
    dict = Uri._normalizeStrictPatternGroups(m.groupdict())
    groups = dict.keys()
    groups.sort()
    for group in groups:
        print "%s: %s" % (group, dict[group])

    This will give you:

    URI: http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right
    URI_reference: http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right#toc
    absoluteURI: http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right
    authority: u:pass@host.foo:8080
    fragment: toc
    hier_part: //u:pass@host.foo:8080/stuff/doc.html?up=down&left=right
    hostport: host.foo:8080
    net_path: //u:pass@host.foo:8080/stuff/doc.html
    port: 8080
    query: up=down&left=right
    scheme: http
    userinfo: u:pass
    """
    r = {}
    for group in matchGroupDict.keys():
        if matchGroupDict[group] is not None:
            newgroup = group
            if group.endswith('_A') or group.endswith('_R'):
                newgroup = group[:-2]
            r[newgroup] = matchGroupDict[group]
    return r


def splitUriRef(uriref):
    """
    Given a valid URI reference as a string, split it into a tuple
    with the following structure, as per RFC 2396 appendix B, with
    the addition of opaque_part to avoid confusion over the meaning
    of 'path':

    (scheme, authority, path, query, fragment, opaque_part)

    All values will be strings (possibly empty) or None. If there
    is an opaque_part, then there will be no authority, path, or
    query.
    """

    _initUriRegex()
    # get scheme, authority, path, query, fragment
    g = splitUriRefPattern.match(uriref).groupdict()
    scheme      = g['scheme']
    authority   = g['authority']
    path        = g['path']
    query       = g['query']
    fragment    = g['fragment']

    # if there is a scheme and the path isn't absolute, then
    # there is an opaque part, not an authority + path + query
    opaque_part = None
    if g['scheme'] and not g['path'].startswith('/'):
        opaque_part = uriref[len(scheme)+1:]
        if fragment:
            opaque_part = opaque_part[:-len(fragment)-1]
        authority = None
        path = None
        query = None

    return (scheme, authority, path, query, fragment, opaque_part)


_setupBeenCompleted = 0
def _initUriRegex():
    """
    Called internally to see if the RegEx needs to be compilied
    """
    global _setupBeenCompleted

    if _setupBeenCompleted:
        return
    _setupBeenCompleted = 1


    URI_REGEX = r"(?:(?:[a-zA-Z][0-9a-zA-Z+\-\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
    FRAGMENT_REGEX = r"(?:#[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
    URIREF_REGEX = "^%s%s$" % (URI_REGEX, FRAGMENT_REGEX)
    URI_REGEX = "^%s$" % (URI_REGEX)

    global uriPattern, uriRefPattern
    uriPattern = re.compile(URI_REGEX)           # lenient checking for URIs
    uriRefPattern = re.compile(URIREF_REGEX)     # lenient checking for URI refs


    #Build all of the pieces of the pattern
    escaped         = r"(?:%[0-9a-fA-F]{2})"
    uric            = r"(?:[;/?:@&=+$,[\]a-zA-Z0-9\-_.!~*'()]|%s)" % escaped # "[" and "]" per RFC 2732
    fragment        = r"(?P<fragment>%s*)" % uric
    query_A         = r"(?P<query_A>%s*)" % uric
    query_R         = r"(?P<query_R>%s*)" % uric
    pchar           = r"(?:[a-zA-Z0-9\-_.!~*'():@&=+$,]|%s)" % escaped
    param           = r"%s*" % pchar
    segment         = r"%s*(?:;%s)*" % (pchar, param)
    path_segments   = r"%s(?:/%s)*" % (segment, segment)
    port            = r"(?:[0-9]*)"
    IPv4address     = r"(?:[0-9]{1,3}\.){3}[0-9]{1,3}"
    hexseq          = r"(?:[0-9a-fA-F]{1,4}(?::[0-9a-fA-F]{1,4})*)"
    hexpart         = r"(?:%s|%s::%s?|::%s?)" % (hexseq, hexseq, hexseq, hexseq)
    IPv6address     = r"%s(?::%s)?" % (hexpart, IPv4address)
    ipv6reference   = r"\[%s\]" % IPv6address
    toplabel        = r"[a-zA-Z](?:[a-zA-Z0-9\-]*[a-zA-Z0-9])?"
    domainlabel     = r"[a-zA-Z0-9](?:[a-zA-Z0-9\-]*[a-zA-Z0-9])?"
    hostname        = r"(?:%s\.)*%s\.?" % (domainlabel, toplabel)
    host            = r"(?:%s|%s|%s)" % (hostname, IPv4address, ipv6reference)
    hostport_A      = r"(?P<hostport_A>%s(?::%s)?)" % (host, port)
    hostport_R      = r"(?P<hostport_R>%s(?::%s)?)" % (host, port)
    userinfo_A      = r"(?P<userinfo_A>(?:[a-zA-Z\-_.!~*'();:&=+$,]|%s)*)" % escaped
    userinfo_R      = r"(?P<userinfo_R>(?:[a-zA-Z\-_.!~*'();:&=+$,]|%s)*)" % escaped
    server_A        = r"(?:(?:%s@)?%s)" % (userinfo_A, hostport_A)
    server_R        = r"(?:(?:%s@)?%s)" % (userinfo_R, hostport_R)
    reg_name        = r"(?:[a-zA-Z\-_.!~*'()$,;:@&=+]|%s)+" % escaped
    authority_A     = r"(?P<authority_A>%s|%s)" % (server_A, reg_name)
    authority_R     = r"(?P<authority_R>%s|%s)" % (server_R, reg_name)
    scheme          = r"(?P<scheme>[a-zA-Z][a-zA-Z0-9+\-.]*)"
    rel_segment     = r"(?:[a-zA-Z\-_.!~*'();@&=+$,]|%s)+" % escaped
    abs_path_noId   = r"(?:/%s)" % path_segments
    abs_path_A      = r"(?P<abs_path_A>/%s)" % path_segments
    abs_path_R      = r"(?P<abs_path_R>/%s)" % path_segments
    rel_path        = r"(?P<rel_path>%s%s?)" % (rel_segment, abs_path_noId)
    net_path_A      = r"(?P<net_path_A>//%s%s?)" % (authority_A, abs_path_noId)
    net_path_R      = r"(?P<net_path_R>//%s%s?)" % (authority_R, abs_path_noId)
    uric_no_slash   = r"(?:[a-zA-Z\-_.!~*'();?:@&=+$,]|%s)" % escaped # no "[" or "]": RFC 2732 oversight?
    opaque_part     = r"(?P<opaque_part>%s%s*)" % (uric_no_slash, uric)
    hier_part       = r"(?P<hier_part>(?:%s|%s)(?:\?%s)?)" % (net_path_A, abs_path_A, query_A)
    relativeURI     = r"(?P<relativeURI>(?:%s|%s|%s)(?:\?%s)?)" % (net_path_R, abs_path_R, rel_path, query_R)
    absoluteURI     = r"(?P<absoluteURI>%s:(?:%s|%s))" % (scheme, hier_part, opaque_part)
    URI             = r"(?P<URI>(?:%s|%s)?)" % (absoluteURI, relativeURI)
    URI_reference   = r"(?P<URI_reference>%s(?:#%s)?)" % (URI, fragment)
    STRICT_URI_PYREGEX = "^%s$" % URI
    STRICT_URIREF_PYREGEX = "^%s$" % URI_reference

    global pyStrictUriPattern,pyStrictUriRefPattern
    pyStrictUriPattern = re.compile(STRICT_URI_PYREGEX)       # strict checking for URIs
    pyStrictUriRefPattern = re.compile(STRICT_URIREF_PYREGEX) # strict checking for URI refs
    
    
    #Build the expressions that are used by the splitUriRef function
    SPLIT_URIREF_REGEX = r"^(?:(?P<scheme>[^:/?#]+):)?(?://(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(?:\?(?P<query>[^#]*))?(?:#(?P<fragment>.*))?$"
    global splitUriRefPattern
    splitUriRefPattern = re.compile(SPLIT_URIREF_REGEX)





#============================================================================
# Extendable resolution functions for URI references
# (needs work)
#
class BaseUriResolver:
    """
    This is the base class that provides a set of functions related to the
    resolution of URIs.
    """
    def normalize(self, uri, base):
        """
        Given a URI reference that is either absolute or relative, and 
        an absolute base URI, this function produces an absolute URI
        reference, effecting the result of the algorithm described in
        section 5.2 of RFC 2396 ("Resolving Relative References to Absolute
        Form"). According to the RFC, the base URI is required to have
        scheme, but for convenience, this function will allow for the
        possibility of the base URI beginning with a '/', in which case
        the 'file:' scheme will be assumed. A UriException is raised if the
        base URI references an unsupported scheme or is otherwise invalid.
        """
        # check that base is valid
        if base is not None and base != '':
            if base[0] == '/':  #assume file: if leading "/" (not ideal)
                base_scheme = 'file'
            else:
                #scheme, netloc, path, params, query, fragment
                base_scheme = urlparse.urlparse(base)[0]
            if base_scheme in ['http', 'ftp', 'file', 'gopher']:
                uri = urllib.basejoin(base, uri)
            # base has unrecognized scheme
            else:
                raise UriException(UriException.INVALID_BASE_URI, base)
            # base is OK
            return uri
        # base is not valid
        else:
            raise UriException(UriException.INVALID_BASE_URI, base)

    def resolve(self, uri, base=None):
        """
        This function takes an absolute or relative URI, and an optional
        base URI, produces a normalized URI using the normalize function,
        attempts to obtain access to an entity representing the resource
        identified by the URI, and returns the entity as a stream. If the
        stream could not be obtained, an exception is raised.
        """
        if base:
            uri = self.normalize(uri, base)
        uri, frag = self.splitFragment(uri)
        if os.access(uri, os.F_OK):
            #Hack because urllib breaks on Windows paths
            # FIXME: This should not be needed because uri
            # should be a URI, not a Windows path.
            stream = open(uri,'rb')
        else:
            stream = urllib.urlopen(uri)
        return stream

    def splitFragment(self, uri):
        frag_split = uri.split("#")
        if len(frag_split) == 2:
            return frag_split[0], frag_split[1]
        else:
            return frag_split[0], ""

    def generate(self, hint=None):
        """
        This function generates and returns a URI.
        The hint is an object that helps decide what to generate.
        The default action is to generate a random UUID URN.
        """
        return 'urn:uuid:'+Uuid.UuidAsString(Uuid.GenerateUuid())


BASIC_RESOLVER = BaseUriResolver()

def OsPathToUri(path, attemptAbsolute=0):
    """
    This function takes an OS path and forces it to a URL of the form
    file:/path/to/file.ext
    Warning: such a file URL might not behave as desired when
    passed to some libraries
    If attemptAbsolute is given and the given path is relative,
    attempt to resolve it to absolute, but as the name indicates,
    this is is just an attempt, and may give the wrong result
    """
    #WARNING: urllib.pathname2url is an undocumented function
    path = urllib.pathname2url(path)
    if path[0] == '/':
        return 'file:' + path
    else:
        if attemptAbsolute:
            return 'file:' + urllib.pathname2url(os.getcwd() + '/') + path
        else:
            return 'file:' + path


def UriToOsPath(uri, attemptAbsolute=0):
    """
    This function takes a file URI and forces it to an OS path
    appropriate for the local system
    If attemptAbsolute is given and the given path is relative,
    attempt to resolve it to absolute, but as the name indicates,
    this is is just an attempt, and may give the wrong result
    """
    scheme, rest = urllib.splittype(uri)
    #Only file URLs are supported
    if scheme and scheme != "file":
        raise Exception("UriToOsPath only supports file type URLs: %s" % uri)
    path = urllib.url2pathname(urllib.splithost(rest)[1])
    return path


#============================================================================

# see the following XML-SIG threads for 'file:' URI resolution discussion
#> http://aspn.activestate.com/ASPN/Mail/Message/xml-sig/614905
#> http://aspn.activestate.com/ASPN/Mail/Message/xml-sig/614913
#> http://mail.python.org/pipermail/xml-sig/2001-February/004572.html

#============================================================================
#
# These are valid absoluteURIs:
#
#  file:/autoexec.bat     (scheme ":" abs_path)
#  file:///autoexec.bat   (scheme ":" net_path)
#
# This one is valid but is not what you'd expect it to be
#
#  file://autoexec.bat    (authority = autoexec.bat, no abs_path)
#
# If you have any more than 3 slashes, it's OK because each path segment
# can be an empty string.
#
# This one is valid too, although everything after 'file:' is
# considered an opaque_part:
#
#   file:etc/passwd
#
# Basically, any file: URI is valid. Good luck resolving them, though.
#
# Jeremy's idea is to not use open() or urllib.urlopen() on Windows;
# instead, use a C function that wraps Windows' generic open function,
# which resolves any path or URI exactly as Explorer would (he thinks)
