import sys, os, signal, socket, select, errno, time, threading

# for Versa queries
from Ft.Rdf.Parsers import Versa
from Ft.Rdf.Parsers.Versa import DataTypes

from Ft.Server.Common import Schema
from Ft.Server.Server import Daemon, Ipc, ConfigParser

# Some functions have *heavy* recursion
sys.setrecursionlimit(10000)


_server_ns = {'ft' : Schema.SCHEMA_NSS + '#'}
_server_versa = Versa.Compile('distribute(ft:server<-ft:type-*, ".",'
                              '".-ft:server.running->*", ".-ft:modified_date->*")')

# Interval, in seconds, between scoreboard maintenance.  During
# each scoreboard maintenance cycle the parent decides if it needs to
# spawn a new child (to meet MinSpareServers requirements), or kill off
# a child (to meet MaxSpareServers requirements).  It will only spawn or
# kill one child per cycle.  Setting this too low will chew cpu.  The
# default is probably sufficient for everyone.  But some people may want
# to raise this on servers which aren't dedicated to httpd and where they
# don't like the httpd waking up each second to see what's going on.
MAINTENANCE_INTERVAL = 1

# This value should prevent resource hogging
MAX_SPAWN_RATE = 32

class Listener:
    def __init__(self, socket, serverConfig):
        self.socket = socket
        self.server = serverConfig
        self.path = serverConfig.path

        # Needed for select()
        self.fileno = socket.fileno
        return

class Worker:
    def __init__(self, config, interval, function, name=None):
        self.config = config
        self.interval = interval
        self.function = function
        self.name = name or function.__name__

        self.ticks = 0

        #Run once is handled elsewhere now
        return

    def tick(self):
        if self.ticks == self.interval:
            self.ticks = 0
            self.run()
        else:
            self.ticks += 1
        return

    def run(self):
        self.config.errorLog.info("running worker '%s'" % self.name)
        repo = None
        try:
            repo = self.config.getRepository()
            self.function(repo)
        except:
            if repo:
                repo.txRollback()
            import traceback
            traceback.print_exc()
        else:
            repo.txCommit()
        return

class DBMaintenanceWorker(Worker):
    def __init__(self, config,when):
        self.config = config
        self._when = when

    def tick(self):
        if time.time() > self._when:
            self.run()
            self._when += 60*60*24

    def run(self):
        self.config.errorLog.info("running worker 'DB Maintenance'")
        repo = None
        try:
            repo = self.config.getRepository()
            driverMod = repo._driver._driverMod
            driverProperties = repo._driver._driverProperties
            repo.txRollback()
            repo = None
            driverMod.Maintain(driverProperties)
        except:
            if repo:
                repo.txRollback()
            import traceback
            traceback.print_exc()
        self.config.errorLog.info("'DB Maintenance' worker done")
        
        
class ControllerBase:

    def __init__(self, config):
        self.config = config

        self.restart_pending = 1
        self.shutdown_pending = 0

        self.listeners = []
        self.daemons = []
        self._idle_spawn_rate = 1

        # Defined here so all daemons will share the same one
        self.accepting_mutex = threading.Lock()
        return

    # -- main execution ----------------------------------------------

    def run(self):
        # Our hooks into the OS
        self.setSignals()

        if self.config.debug:
            # Let Control-C do a clean shutdown
            signal.signal(signal.SIGINT, self.shutdown)

        while self.restart_pending:

            self.config.readConfig()
            self.config.openLogs()

            self.errorLog = self.config.errorLog
            self.config.savePid()

            workers = self.getWorkers()

            for slot in range(self.config.daemons_to_start):
                daemon = Daemon.Daemon(self.accepting_mutex, self.listeners,
                                       self.errorLog)
                daemon.start()
                self.daemons.append(daemon)

            self.errorLog.notice("%s configured -- resuming normal "
                                 "operations" % self.config.ident)

            self.restart_pending = self.shutdown_pending = 0

            while not self.restart_pending and not self.shutdown_pending:
                self.waitOrTimeout(MAINTENANCE_INTERVAL)
                self.idleMaintenance()

                # Update the workers internal counters; this might cause
                # them to execute as well
                for worker in workers:
                    worker.tick()

            if self.shutdown_pending:
                self.errorLog.notice('shutting down')

            self.reclaimChildren()
            
            # debug mode doesn't have the concept of restarting
            if self.shutdown_pending or self.debug:
                # Cleanup pid file on normal shutdown
                self.config.removePid()
                break

            # we've been told to restart
            self.errorLog.notice('doing graceful restart')
            continue

        self.errorLog.notice('process %d exitting' % os.getpid())
        return 0

    # -- signal handling ---------------------------------------------

    def setSignals(self):
        raise NotImplementedError

    def waitOrTimeout(self, timeout):
        raise NotImplementedError

    def killDaemon(self, daemon, nice):
        raise NotImplementedError

    def shutdown(self, *ignored):
        # the varargs is to allow this as a signal handler
        self.shutdown_pending = 1
        return

    def restart(self, *ignored):
        # the varargs is to allow this as a signal handler
        self.restart_pending = 1
        return

    # -- connection structures and accouting -------------------------

    def makeSocket(self, serverConfig):
        """
        Creates the socket for this server.
        """
        # Create a human readable version of the address for errors
        # and a tuple for use with bind()
        if self.config.bindAddress:
            bind_addr = (self.config.bindAddress, serverConfig.port)
            addr = 'address %s port %d' % bind_addr
        else:
            # The empty string represents INADDR_ANY
            bind_addr = ('', serverConfig.port)
            addr = 'port %d' % serverConfig.port

        self.errorLog.debug('creating socket for %s' % addr)
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except:
            msg = 'failed to get a socket for %s' % addr
            self.errorLog.critical(msg)
            return None

        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except:
            msg = 'for %s, setsockopt: SO_REUSEADDR' % addr
            self.errorLog.critical(msg)
            return None

        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        except:
            msg = 'for %s, setsockopt: SO_KEEPALIVE' % addr
            self.errorLog.critical(msg)
            return None

        # The Nagle algorithm says that we should delay sending partial
        # packets in hopes of getting more data.  We don't want to do
        # this; we are not telnet.  There are bad interactions between
        # persistent connections and Nagle's algorithm that have very severe
        # performance penalties.  (Failing to disable Nagle is not much of a
        # problem with simple HTTP.)
        #
        # In spite of these problems, failure here is not a shooting offense.
        try:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        except:
            msg = 'for %s, setsockopt: TCP_NODELAY'
            self.errorLog.warning(msg)

        try:
            sock.bind(bind_addr)
        except:
            msg = 'could not bind to %s' % addr
            self.errorLog.critical(msg)
            return None

        try:
            # If the socket is of type AF_INET, and the backlog argument is
            # greater than the constant SO_MAXCONN (128 in 2.0.23), it is
            # silently truncated to SO_MAXCONN.  For portable applications,
            # don't rely on this value since BSD (and some BSD-derived systems)
            # limit the backlog to 5.
            sock.listen(5)
        except:
            msg = 'unable to listen for connections on %s' % addr
            self.errorLog.critical(msg)
            return None

        return sock

    # -- daemon control ----------------------------------------------

    def reclaimChildren(self):
        for daemon in self.daemons:
            # Prevent processing any more requests
            daemon.active = 0

        self.errorLog.info('waiting for daemons to exit...')

        waittime = 0.016  # 16ms
        for tries in range(13):
            # don't want to hold up progess any more than
            # necessary, but we need to allow children a few moments
            # to exit. Set delay with an exponential backoff.
            time.sleep(waittime)
            waittime *= 4

            not_dead_yet = 0
            for daemon in self.daemons:
                if daemon.isAlive():
                    not_dead_yet += 1
                    if tries == 3: # 344ms
                        # perhaps it missed the first attempt
                        daemon.active = 0
                        waittime = 0.016
                    elif tries == 8: # 6 secs
                        self.errorLog.warning("%s did not exit, sending "
                                              "interrupt" % daemon)
                        self.killDaemon(daemon, 1)
                        waittime = 0.016
                    elif tries == 9: # 6 secs
                        self.errorLog.error("%s still did not exit, sending "
                                            "termination" % daemon)
                        self.killDaemon(daemon, 0)
                        waittime = 0.016
                    if tries == 12: # 7.4 secs
                        self.errorLog.error("could not make %s exit, attempting"
                                            " to continue anyway" % daemon)

            if not not_dead_yet:
                # nothing left to wait for
                break
        return

    def idleMaintenance(self):
        inactive = []
        idle_count = 0

        for daemon in self.daemons:
            if not daemon.active:
                inactive.append(daemon)
            elif daemon.ready:
                idle_count += 1

        # Remove any dead threads:
        if inactive:
            self.errorLog.notice('Purging %d unused daemons' % len(inactive))
            for daemon in inactive:
                self.daemons.remove(daemon)

        if idle_count > self.config.daemons_max_free:
            # kill off one child...let it die gracefully
            # always kill the highest numbered child if we have to...
            # no really well thought out reason ... other than observing
            # the server behaviour under linux where lower numbered children
            # tend to service more hits (and hence are more likely to have
            # their data in cpu caches).
            if self.daemons:
                # We might not have any current children
                self.daemons[-1].active = 0
            self._idle_spawn_rate = 1

        elif idle_count < self.config.daemons_min_free:
            if len(self.daemons) > self.config.max_daemons:
                logger.error('%s: reached MaxClients setting' % self.path)
                self._idle_spawn_rate = 1
            else:
                if self._idle_spawn_rate >= 8:
                    self.errorLog.info(
                        '%s: server seems busy, (you may need to increase '
                        'Min/MaxSpareServers), creating %d children, there '
                        'are %d idle, and %d total children' % (
                        self.path, self._idle_spawn_rate, idle_count,
                        len(self.daemons)))

                self.errorLog.notice('spawning %d new children' %
                                     self._idle_spawn_rate)
                for slot in range(self._idle_spawn_rate):
                    daemon = Daemon.Daemon(self.accepting_mutex, self.listeners,
                                           self.errorLog)
                    daemon.start()
                    self.daemons.append(daemon)

                # the next time around we want to spawn twice as many if this
                # wasn't good enough
                if self._idle_spawn_rate < MAX_SPAWN_RATE:
                    self._idle_spawn_rate *= 2
        else:
            self._idle_spawn_rate = 1
        return

    # -- worker threads ----------------------------------------------

    def getWorkers(self):
        workers = []

        # Dynamic configuration update
        interval = self.config.properties['DynamicReloadInterval']
        worker = Worker(self.config, interval, self._check_config)
        worker.run()
        if interval > 0:
            workers.append(worker)
            

        # Temporary file purge thread
        interval = self.config.properties['TemporaryReapInterval']
        worker = Worker(self.config, interval, self._purge_temp)
        worker.run()
        if interval > 0:
            workers.append(worker)

        # XSLT Cron
        interval = self.config.properties['XsltStrobeInterval']
        worker = Worker(self.config, interval, self._xslt_strobe)
        worker.run()
        if interval > 0:
            workers.append(worker)

        # DB Maint.
        interval = self.config.properties['DBMaintenanceTime']
        if interval > 0:
            worker = DBMaintenanceWorker(self.config, interval)
            workers.append(worker)

        return workers

    def _check_config(self, repo):
        # Determine all available servers (via schema) and the
        # "running" and "last-modified-date" for them
        context = Versa.CreateContext(model=repo.getModel(),
                                      nsMapping=_server_ns)
        result = _server_versa.evaluate(context)

        if not result:
            # No servers defined
            return

        status = {}
        for name, running, modified in result:
            running = len(running) and int(min(DataTypes.ToList(running))) or 0
            modified = max(DataTypes.ToList(modified))
            status[str(name)] = (running, modified)

        unused = []
        # Process currently running servers
        for listener in self.listeners:
            if not status.has_key(listener.path):
                # The configuration file has been removed
                unused.append(listener)
                continue

            running, modified_date = status[listener.path]
            if not running:
                # The server should be stopped
                unused.append(listener)
                del status[listener.path]
            elif modified_date > listener.server.modifiedDate:
                # The configuration file has changed
                # Remove the current listener and recreate it with
                # the new configuration
                unused.append(listener)
            else:
                # By removing it from status, no changes will be made
                del status[listener.path]

        for listener in unused:
            self.listeners.remove(listener)

        # Servers remaining in status are considered to be new
        if status:
            items = status.items()
            items = [ (path, date) for path, (start, date) in items if start ]
            
            parser = ConfigParser.ConfigParser()
            try:
                repo = self.config.getRepository()
            except:
                # This is an extremely critical error, just die
                self.errorLog.emergency('Unable to connect to repository')
                self.shutdown()
                return

            try:
                for path, date in items:
                    try:
                        resource = repo.fetchResource(path)
                        config = parser.readConfig(path, resource.getContent())
                        # Prevent handlers from corrupting the configuration
                        config.finalize(date, self.config.properties.copy())
                    except Exception, error:
                        self.errorLog.error('Unable to read configuration '
                                            'resource %s: %s' % (path, error))
                        continue

                    socket = self.makeSocket(config)
                    if socket:
                        config.openLogs(self.config)
                        listener = Listener(socket, config)
                        self.listeners.append(listener)
            finally:
                repo.txCommit()
        return

    def _purge_temp(self, repo):
        repo.purgeTemporaryUris()
        return

    def _xslt_strobe(self, repo):
        repo.runXsltStrobe()
        return


class PosixController(ControllerBase):

    def setSignals(self):
        signal.signal(signal.SIGTERM, self.shutdown)
        signal.signal(signal.SIGHUP, self.restart)
        return

    def waitOrTimeout(self, timeout):
        time.sleep(timeout)
        return

    def killDaemon(self, daemon, nice):
        pid = daemon.pid
        if nice == 1:
            # try interrupt
            os.kill(pid, signal.SIGINT)
        else:
            # kill it!
            os.kill(pid, signal.SIGTERM)
        return

class WindowsController(ControllerBase):

    def setSignals(self):
        # Windows doesn't really use signals, therefore we'll use events
        # that mimic how signals are used
        shutdown = Ipc.Event('ap%dshutdown' % os.getpid())
        shutdown.handle = self.shutdown
        restart = Ipc.Event('ap%drestart' % os.getpid())
        restart.handle = self.restart
        self._signals = (shutdown, restart)
        return

    def waitOrTimeout(self, timeout):
        set = Ipc.WaitEvents(self._signals, timeout)
        for event in set:
            event.clear()
            event.handle()
        return

    def killDaemon(self, daemon, nice):
        # Windows can exit the program with active threads
        return

if sys.platform == 'win32':
    Controller = WindowsController

elif os.name == 'posix':
    Controller = PosixController

else:
    raise ImportError("I don't know how to control servers on this platform!")


if __name__ == '__main__':
    import GlobalConfig
    username, password, core = eval(os.environ['FTSS_AGENT'])
    filename = os.environ['FTSERVER_CONFIG_FILE']
    config = GlobalConfig.GlobalConfig(username, password, core, filename)
    controller = Controller(config)
    status = controller.run()
    sys.exit(status)

