import sys, signal, time, os
#import string

from Ft.Server.Server import Ipc, Controller
#from Ft.Server.Server.Lib import ConfigFile

PID_RUNNING = 0
PID_DEAD = 1
PID_STOPPED = 2
PID_NOT_OWNER = 3

# This get mapped to the appropriate platform launcher
# (will be set later)
Launcher = None

def RemovePid(pidFile):
    try:
        if os.path.exists(pidFile):
            os.remove(pidFile)
    except:
        pass
    return

def WaitPid(pidFile, seconds=30):
    timeout = time.time() + seconds
    while os.path.exists(pidFile) and time.time() < timeout:
        time.sleep(0.1)
    return os.path.exists(pidFile)

class LauncherBase:
    def __init__(self, username, password, properties):
        # username, password will provided from the commands authentication
        # properties, filename refer to the configuration used in the command

        self.pidFile = properties['PidFile']
        if not os.path.exists(os.path.dirname(self.pidFile)):
            raise Exception("Unable to use pid file '%s', path is invalid" % self.pidFile)

        logfile = properties['LogFile']
        if not os.path.exists(os.path.dirname(logfile)):
            raise Exception("Unable to use log file '%s', path is invalid" % logfile)

        config = properties.get('Server', {'ServerType' : 'Python'})
        if config['ServerType'] == 'Apache':
            self.name = 'apache'
            executable = os.path.abspath(config['Executable'])
            self.chroot, basename = os.path.split(executable)
            #FIXME - write out the config file for apache to use
            # -X means run in foreground (debug mode)
            args = (os.path.basename(executable), '-X')

        elif config['ServerType'] == 'Python':
            self.name = 'controller'
            executable = sys.executable
            self.chroot, basename = os.path.split(Controller.__file__)
            # When running .pyc files, Python 2.1 has debugging output
            # and will probably cause some confusion for users
            if basename[-4:] == '.pyc':
                basename = basename[:-1]
            args = (os.path.basename(executable), basename)
        else:
            raise Exception('Unknown server configuration')

        # Create the sub-process environment
        # os.environ is a UserDict, we need a real dictionary
        environ = os.environ.data.copy()

        agent_data = (username, password, properties['CoreId'])
        environ['FTSS_AGENT'] = repr(agent_data)
        environ['FTSERVER_CONFIG_FILE'] = properties['ConfigFile']

        self.process = Ipc.Process(executable, args, environ)
        return

    def info(self, msg):
        sys.stdout.write(msg + '\n')
        return

    def error(self, msg):
        msg = '[ERROR] ' + msg
        sys.stderr.write(msg + '\n')
        raise Exception(msg)

    def _remove_pid(self):
        if os.path.exists(self.pidFile):
            try:
                os.remove(self.pidFile)
            except Exception, error:
                sys.stderr.write("Unable to remove pid file '%s': %s\n" %
                                 (self.pidFile, str(error)))
        return

    def stop(self):
        (pid, status) = self.status()
        self._remove_pid()
        if status == PID_RUNNING:
            self.info('Stopping %s' % self.name)
            self._stop(pid)
        elif status == PID_DEAD:
            self.info('Removing stale pid file (pid %d)' % pid)
        else:
            self.info('%s is not running' % self.name)
        return 0

    def start(self):
        (pid, status) = self.status()
        if status == PID_RUNNING:
            self.info('%s is already running' % self.name)
            return
        # Make sure there is a clean slate
        self._remove_pid()
        cwd = os.getcwd()
        try:
            os.chdir(self.chroot)
            pid = self.process.start()
            exited, code = self.process.wait(0.5)
            while not exited and not os.path.exists(self.pidFile):
                exited, code = self.process.wait(0.5)
        finally:
            os.chdir(cwd)

        if not exited:
            self.info('%s started (pid %s)' % (self.name, pid))
        else:
            self.info('Error starting %s' % self.name)
            self.info('FIXME - no extended status available')
        return pid

    def restart(self):
        (pid, status) = self.status()
        if status == PID_RUNNING:
            self.info('Restarting %s...' % self.name)
            self._stop(pid)
        if WaitPid(self.pidFile):
            self.error('Unable to stop %s.  If you know it\'s not already running, stop and then start again' % self.name)
            pid = 0
        else:
            pid = self.start()
        return pid

    def reload(self):
        (pid, status) = self.status()
        if status == PID_RUNNING:
            self.info('Restarting %s...' % self.name)
            self._reload(pid)
        else:
            self.info('%s not running' % self.name)
        return pid

    def status(self):
        pid, status = 0, PID_STOPPED

        if os.path.exists(self.pidFile):
            try:
                fd = open(self.pidFile)
            except Exception, error:
                sys.stderr.write("Unable to open pid file '%s': %s\n" %
                                 (self.pidFile, str(error)))
                return (pid, status)

            try:
                content = fd.read()
            except Exception, error:
                sys.stderr.write("Unable to read pid file '%s': %s\n" %
                                 (self.pidFile, str(error)))
                try:
                    fd.close()
                except:
                    sys.stderr.write("Error closing file descriptor\n")
                return (pid, status)

            try:
                fd.close()
            except:
                sys.stderr.write("Error closing file descriptor\n")

            try:
                pid = int(content.strip())
                status = PID_RUNNING
            except:
                sys.stderr.write("Invalid pid file '%s'\n" % self.pidFile)
        return (pid, status)

    def _stop(self, pid):
        return


class PosixLauncher(LauncherBase):

    def _stop(self, pid):
        try:
            os.kill(pid, signal.SIGTERM)
        except:
            pass
        return

    def _reload(self, pid):
        try:
            os.kill(pid, signal.SIGHUP)
        except:
            pass
        return


class WindowsLauncher(LauncherBase):

    def _stop(self, pid):
        event = Ipc.Event('ap%dshutdown' % pid)
        event.set()
        return

    def _reload(self, pid):
        event = Ipc.Event('ap%drestart' % pid)
        event.set()
        return


if sys.platform == 'win32':
    Launcher = WindowsLauncher

elif os.name == 'posix':
    Launcher = PosixLauncher

else:
    raise SystemExit("I don't know how to start servers this platform!")
