import StreamWriter

class XmlPrinter:

    def __init__(self, stream, encoding):
        """
        Creates a Printer instance.

        stream must be a file-like object open for writing (binary) data.
        encoding specifies the encoding which is to be used for the stream.
        """
        self.stream = sw = StreamWriter.StreamWriter(stream, encoding)
        self.encoding = encoding
        self.writeAscii = sw.writeAscii
        self.writeEncode = sw.writeEncode
        self.writeEscape = sw.writeEscape
        self.inElement = 0
        return

    def startDocument(self, version='1.0', standalone=None):
        self.writeAscii('<?xml version="%s" encoding="%s"' % (version,
                                                               self.encoding))
        if standalone is not None:
            self.writeAscii(' standalone="%s"' % standalone)

        self.writeAscii('?>\n')
        return

    def endDocument(self):
        if self.inElement:
            # No element content, use minimized form
            self.writeAscii('/>')
        return
    
    def doctype(self, name, publicId, systemId):
        if publicId and systemId:
            self.writeAscii('<!DOCTYPE ')
            self.writeEncode(name, 'document type name')
            self.writeAscii(' PUBLIC "')
            self.writeEncode(publicId, 'document type public-id')
            self.writeAscii('" "')
            self.writeEncode(systemId, 'document type system-id')
            self.writeAscii('">\n')
        elif systemId:
            self.writeAscii('<!DOCTYPE ')
            self.writeEncode(name, 'document type name')
            self.writeAscii(' SYSTEM "')
            self.writeEncode(systemId, 'document type system-id')
            self.writeAscii('">\n')
        return

    def startElement(self, tagName, namespaces, attributes):
        if self.inElement:
            # Close current start tag
            self.writeAscii('>')
        else:
            self.inElement = 1

        self.writeAscii('<')
        self.writeEncode(tagName, 'start-tag name')
        
        # Write the namespaces
        for prefix, uri in namespaces.items():
            if prefix:
                self.attribute(tagName, u"xmlns:" + prefix, uri)
            else:
                self.attribute(tagName, u"xmlns", uri)

        # Now the attributes
        for name, value in attributes.items():
            self.attribute(tagName, name, value)
        return

    def endElement(self, tagName):
        if self.inElement:
            # No element content, use minimized form
            self.writeAscii('/>')
            self.inElement = 0
        else:
            self.writeAscii('</')
            self.writeEncode(tagName, 'end-tag name')
            self.writeAscii('>')
        return
    
    # elementName is only need for HTML output
    def attribute(self, elementName, name, value):
        self.writeAscii(" ")
        self.writeEncode(name, 'attribute name')
        self.writeAscii("=")

        # Replace characters illegal in attribute values
        # Wrap the value with appropriate quoting in accordance with
        # DOM Level 3 Load and Save:
        # 1. Attributes not containing quotes are serialized in quotes.
        # 2. Attributes containing quotes but no apostrophes are serialized
        #    in apostrophes.
        # 3. Attributes containing both forms of quotes are serialized in
        #    quotes, with quotes within the value represented by the
        #    predefined entity &quot;.
        if (u'"' in value) and (u"'" not in value):
            # Use apostrophes (#2)
            entitymap = self.attrEntitiesApos
            quote = "'"
        else:
            # Use quotes (#1 and #3)
            entitymap = self.attrEntitiesQuot
            quote = '"'

        self.writeAscii(quote)
        self.writeEscape(value, entitymap)
        self.writeAscii(quote)
        return

    def text(self, text, disableEscaping=0):
        if self.inElement:
            self.writeAscii('>')
            self.inElement = 0

        if disableEscaping:
            # Try to write the raw encoded string (may throw exception)
            self.writeEncode(text, 'text')
        else:
            self.writeEscape(text, self.textEntities)
        return
    
    def cdataSection(self, data):
        if self.inElement:
            self.writeAscii('>')
            self.inElement = 0

        sections = data.split(u']]>')
        self.writeAscii('<![CDATA[')
        self.writeEncode(sections[0], 'CDATA section')
        for section in sections[1:]:
            self.writeAscii(']]]]><![CDATA[>')
            self.writeEncode(section, 'CDATA section')
        self.writeAscii(']]>')
        return

    def processingInstruction(self, target, data):
        if self.inElement:
            self.writeAscii('>')
            self.inElement = 0

        self.writeAscii('<?')
        self.writeEncode(target, 'processing instruction target')
        if data:
            self.writeAscii(' ')
            self.writeEncode(data, 'processing instruction data')
        self.writeAscii('?>')
        return

    def comment(self, data):
        if self.inElement:
            self.writeAscii('>')
            self.inElement = 0

        self.writeAscii("<!--")
        self.writeEncode(data, 'comment')
        self.writeAscii("-->")
        return

    textEntities = StreamWriter.EntityMap({'<' : '&lt;',
                                           '>' : '&gt;',
                                           '&' : '&amp;',
                                           })

    attrEntitiesQuot = StreamWriter.EntityMap({'<' : '&lt;',
                                               '>' : '&gt;',
                                               '&' : '&amp;',
                                               '\n' : '&#10;',
                                               '"' : '&quot;',
                                               })

    attrEntitiesApos = StreamWriter.EntityMap({'<' : '&lt;',
                                               '>' : '&gt;',
                                               '&' : '&amp;',
                                               '\n' : '&#10;',
                                               "'" : '&apos;',
                                               })

