########################################################################
#
# File Name:            FtElements.py
#
# Documentation:        http://docs.4suite.org/4XSLT/FtElements.py.html
#
"""
Fourthought proprietary extension elements
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import types
from Ft.Xml import EMPTY_NAMESPACE
from Ft.Xml.Xslt import XSL_NAMESPACE, XsltElement, XsltException, Error
from Ft.Xml.Xslt import ContentInfo, AttributeInfo
from Ft.Xml.Xslt import ApplyTemplatesElement
from Ft.Xml.Xslt import OutputParameters
from Ft.Xml.XPath import Conversions
from Ft.Xml.XPath import FT_EXT_NAMESPACE, FT_OLD_EXT_NAMESPACE

RESERVED_NAMESPACE = 'http://xmlns.4suite.org/reserved'

class FtApplyTemplates(ApplyTemplatesElement.ApplyTemplatesElement):
    """
    The f:apply-templates element is a simple extension of xsl:apply-templates
    which allows you to compute the mode dynamically (as an attribute value
    template)
    """
    legalAttrs = ApplyTemplatesElement.ApplyTemplatesElement.legalAttrs.copy()
    legalAttrs['mode'] = AttributeInfo.QNameAvt(description="The mode to be used for template application.  In this variation the mode is an AVT and thus can be computedd at run time.")

    def _instantiate_mode(self, context):
        return self._mode.evaluate(context)


class FtOutputElement(XsltElement):
    """
    f:output is a simple extension of xsl:output which allows you to
    compute the output  parameters dynamically (as attribute value
    templates)
    """

    content = ContentInfo.Template
    legalAttrs = {
        'method' : AttributeInfo.QNameAvt(),
        'version' : AttributeInfo.NMTokenAvt(),
        'encoding' : AttributeInfo.StringAvt(),
        'omit-xml-declaration' : AttributeInfo.YesNoAvt(),
        'standalone' : AttributeInfo.YesNoAvt(),
        'doctype-public' : AttributeInfo.StringAvt(),
        'doctype-system' : AttributeInfo.StringAvt(),
        'cdata-section-elements' : AttributeInfo.QNamesAvt(),
        'indent' : AttributeInfo.YesNoAvt(),
        'media-type' : AttributeInfo.StringAvt(),
        }

    def __init__(self, root, namespaceUri, localName, baseUri):
        XsltElement.__init__(self, root, namespaceUri, localName, baseUri)
        self._output_parameters = OutputParameters.OutputParameters()
        return

    def instantiate(self, context, processor):
        context.processorNss = self.namespaces
        context.currentInstruction = self

        # this uses attributes directly from self
        self._output_parameters.avtParse(self, context)

        processor.addHandler(self._output_parameters,
                             processor.writers[-1]._stream)
        try:
            for child in self.children:
                context = child.instantiate(context, processor)[0]
        finally:
            processor.removeHandler()

        return (context,)


class AssignElement(XsltElement):
    """The f:assign element is used to create a new variable, or change the
    value of a local or global variable that has previously been declared
    using xsl:variable or xsl:param.

    As with xsl:variable, the name of the variable is given in the mandatory
    name attribute, and the new value may be given either by an expression in
    the select attribute, or by instantiating the content of the f:assign
    element.

    If no select attribute is given, then a body-as-ns attribute may be used
    to indicate whether to assign the variable to the contents as a node-set
    (value 'yes') or as a result tree fragment (default, or value 'no').
    In either case, be aware that the node-set or result tree fragment will
    have a root node.

    Note that reassignment of variables is generally never actually needed.
    Before using f:assign, read the XSL FAQ or ask on xsl-list if there is a
    better, more portable way to solve your problem.

    XSLT is designed as a language that is free of side-effects, which is
    why assignment is not allowed and variables have very specific scope.
    When variable assignment is allowed, certain optimizations in the XSLT
    processor become impossible. Also, there are some circumstances in which
    the order of execution may not be quite what you expect, in which case
    f:assign may show anomalous behavior.

    That said, f:assign can be a convenient way, to create, in a single step,
    a variable binding to a node-set that was generated using the f:assign
    element's content. The only other way to do that with xsl:variable, in
    XSLT 1.0, would be to create a result tree fragment and bind a second
    variable to the result of calling the EXSLT node-set() function on the
    first."""

    legalAttrs = {
        'name' : AttributeInfo.QName(required=1),
        'select' : AttributeInfo.Expression(),
        'body-as-ns' : AttributeInfo.YesNoAvt(default='no'),
        }

    def instantiate(self, context, processor):
        context.processorNss = self.namespaces
        context.currentInstruction = self
        if self._select:
            result = self._select.evaluate(context)
        else:
            processor.pushResult()
            try:
                for child in self.children:
                    context = child.instantiate(context, processor)[0]
            finally:
                result = processor.popResult()

            if self._body_as_ns.evaluate(context):
                result = [result]

        context.varBindings[self._name] = result
        context.processor.stylesheet.globalVars[self._name] = result
        return (context,)


class DumpKeysElement(XsltElement):
    """
    f:dump-keys dumps to the output the items in the XSLT keys
    currently defined, for the document that owns the context node.
    This is useful for debugging.
    By default, the key data is presented in an XML format such as:

    
    """

    legalAttrs = {
        'raw' : AttributeInfo.YesNoAvt(default='no', description="Present keys in a compact non-XML format"),
        }

    def instantiate(self, context, processor):
        doc = context.node.ownerDocument or context.node

        try:
            xkeys = processor.keys[doc]
        except KeyError:
            xkeys = {}

        if self._raw.evaluate(context) == 'yes':
            processor.xslMessage(repr(xkeys))
        else:
            from Ft.Xml.Xslt.CopyOfElement import CopyNode
            processor.writers[-1].startElement(u'zz:KeyDump', RESERVED_NAMESPACE)
            for k, v in xkeys.items():
                processor.writers[-1].startElement(u'zz:Key', RESERVED_NAMESPACE)
                processor.writers[-1].attribute(u'name', k[1], EMPTY_NAMESPACE)
                for kk, vv in v.items():
                    processor.writers[-1].startElement(u'zz:MatchSet', RESERVED_NAMESPACE)
                    processor.writers[-1].attribute(u'value', kk, EMPTY_NAMESPACE)
                    for node in vv:
                        CopyNode(processor, node)
                processor.writers[-1].endElement(u'zz:Key')
            processor.writers[-1].endElement(u'zz:KeyDump')
        return (context,)


class DumpVarsElement(XsltElement):
    """
    f:dump-vars dumps the current variables and parameters in scope
    to the output.  This is useful for debugging.
    By default, the variables are presented in an XML format such as:

    
    
    """
    legalAttrs = {
        'raw' : AttributeInfo.YesNoAvt(default='no', description="Present variables and values in a compact non-XML format"),
        }

    def instantiate(self, context, processor):
        doc = context.node.ownerDocument or context.node
        if self._raw.evaluate(context):
            processor.xslMessage(repr(context.varBindings))
        else:
            from Ft.Xml.XPath import g_xpathPrimitiveTypes
            from Ft.Xml.Xslt.CopyOfElement import CopyNode
            processor.writers[-1].startElement(u'zz:VarDump', RESERVED_NAMESPACE)
            for k, v in context.varBindings.items():
                processor.writers[-1].startElement(u'zz:Var', RESERVED_NAMESPACE)
                #FIXME: should try to join back prefix to var name
                processor.writers[-1].attribute(u'name', k[1], EMPTY_NAMESPACE)
                if type(v) == type([]):
                    # NOTE - this must be before the primitive check due to
                    #  the fact that a node-set is a primitive type
                    for node in v:
                        CopyNode(processor, node)
                elif type(v) in g_xpathPrimitiveTypes:
                    processor.writers[-1].text(Conversions.StringValue(v))
                elif hasattr(v, 'nodeType'):
                    CopyNode(processor, v)
                processor.writers[-1].endElement(u'zz:Var')
            processor.writers[-1].endElement(u'zz:VarDump')
        return (context,)


class DumpContextElement(XsltElement):
    """
    f:dump-context dumps the context information
    to the output.  This is useful for debugging.
    By default, the content is presented in an XML format such as:
    
    """

    legalAttrs = {}

    def instantiate(self, context, processor):
        doc = context.node

        from cStringIO import StringIO
        from Ft.Xml.Xslt.CopyOfElement import CopyNode
        stream = StringIO()

        writer = processor.writers[-1]
        processor.addHandler(writer._outputParams, stream)
        for child in doc.childNodes:
            CopyNode(processor,child)
        processor.removeHandler()
        text = stream.getvalue()
        processor.xslMessage(text)
        return (context,)


class ReplaceElement(XsltElement):
    """
    f:replace performs a search and replace on a string, placing the results
    in the output.  The content is treated as a template.  The string value
    of the output from this template is the replacement string.
    All instances of the string given by the 'substring' attribute
    are replaced with the replacement string.
    """

    legalAttrs = {
        'string' : AttributeInfo.StringExpression(description="The string to be processed.  If not given, the string value of the context node is used."),
        'substring' : AttributeInfo.StringExpression(required=1, description="The sub-string to be replaced."),
        }

    def instantiate(self, context, processor):
        context.processorNss = self.namespaces
        context.currentInstruction = self

        if self._string:
            value = self._string.evaluate(context)
        else:
            value = context.node

        string_ = Conversions.StringValue(value)
        substring = Conversions.StringValue(self._substring.evaluate(context))
        for chunk in string_.split(substring):
            processor.writers[-1].text(chunk)
            for child in self.children:
                context = child.instantiate(context, processor)[0]

        return (context,)


class MsgControlElement(XsltElement):
    """
    f:msg-control is a special utility that allows one to control the display of XSL messages
    The 'suppress' attribute allows you to disable display of all messages.
    """

    legalAttrs = {
        'suppress' : AttributeInfo.YesNoAvt(default='no', description="Disable display of all XSLT messages."),
        }

    def instantiate(self, context, processor):
        processor.messageControl(self._suppress.evaluate(context))
        return (context,)


class CreateIndexElement(XsltElement):
    """
    f:create-index allows one to create an arbitrary key at run time using
    any node data.  It is similar to xsl:key, except that it is computed
    on demand at run-time, and uses an XPath selection rather than an XSLT
    match, which gives more flexibility over what is indexed.

    These keys can be accessed using the regular XSLT function key().
    """
    
    legalAttrs = {
        'name' : AttributeInfo.StringAvt(required=1, description='The name of the key to create'),
        'select' : AttributeInfo.Expression(required=1, description='Selects which nodes are to be indexed'),
        'use' : AttributeInfo.Expression(required=1, description='The expression that computes the index key value for each node'),
        }

    def instantiate(self, context, processor):
        selected = self._select.evaluate(context)
        name = self._name.evaluate(context)
        index = {}
        if not isinstance(selected, types.ListType):
            return (context,)
        state = context.copy()
        size = len(selected)
        pos = 1
        for node in selected:
            context.node, context.position, context.size = node, pos, size
            context.currentNode = node
            key = Conversions.StringValue(self._use.evaluate(context))
            if not index.has_key(key):
                index[key] = []
            index[key].append(node)
            pos += 1
        if not processor.extensionParams.has_key((FT_EXT_NAMESPACE, 'indices')):
            processor.extensionParams[(FT_EXT_NAMESPACE, 'indices')] = {}
        processor.extensionParams[(FT_EXT_NAMESPACE, 'indices')][name] = index
        context.set(state)
        return (context,)


ExtNamespaces = {
    FT_EXT_NAMESPACE : 'f',
    }

ExtElements = {
    (FT_EXT_NAMESPACE, 'apply-templates'): FtApplyTemplates,
    (FT_EXT_NAMESPACE, 'output'): FtOutputElement,
    (FT_EXT_NAMESPACE, 'dump-keys'): DumpKeysElement,
    (FT_EXT_NAMESPACE, 'dump-vars'): DumpVarsElement,
    (FT_EXT_NAMESPACE, 'assign'): AssignElement,
    (FT_EXT_NAMESPACE, 'replace'): ReplaceElement,
    (FT_EXT_NAMESPACE, 'dump-context'): DumpContextElement,
    (FT_EXT_NAMESPACE, 'message-control'): MsgControlElement,
    (FT_EXT_NAMESPACE, 'create-index'): CreateIndexElement,
    }

