########################################################################
#
# File Name:   BuiltInExtFunctions.py
#
# Docs:        http://docs.4suite.org/XPath/BuiltInExtFunctions.py.html
#
"""
4XSLT specific extension functions (i.e. ones that create a node set)
WWW: http://4suite.org/XSLT        e-mail: support@4suite.org

Copyright (c) 2000-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

#Use this module only for functions that generate new node sets
#or for whatever reason need a processor object (e.g. FtKey needs
#a processor to read the key defs)
#Place others in Ft.Xml.XPath.BuiltInExtFunctions.py

import re
from Ft.Xml import SplitQName
#from Ft.Xml.XPath import RuntimeException
from Ft.Xml.XPath import FT_EXT_NAMESPACE
from Ft.Xml.XPath import CoreFunctions, Conversions
from Ft.Xml.Xslt import XsltFunctions, XsltRuntimeException, Error


def FtKey(context, qname, keyList, realContextNode=None):
    """
    Query an XSLT key, but allow the user to override the context node for
    purposes of determining which source document to check.
    realContextNode is a node set from which the first node is extracted
    And used as the context node
    """
    if realContextNode:
        orig_node = context.node
        context.node = realContextNode[0]
        result = XsltFunctions.Key(context, qname, keyList)
        context.node = orig_node
    else:
        result = XsltFunctions.Key(context, qname, keyList)
    return result


def SearchRe(context, pattern, arg=None):
    """Do a regular expression search against the argument (i.e. get all matches)"""
    if not arg:
        arg = context.node
    arg = Conversions.StringValue(arg)
    processor = context.processor
    matches_nodeset = []
    compiled = re.compile(pattern)
    match = compiled.search(arg)
    while match:
        processor.pushResult()
        processor.writers[-1].startElement('Match')
        # .groups() return empty tuple when the pattern did not do grouping
        groups = match.groups() or (match.group(),)
        for group in groups:
            processor.writers[-1].startElement('Group')
            group and processor.writers[-1].text(group)
            processor.writers[-1].endElement('Group')
        processor.writers[-1].endElement('Match')
        rtf = processor.popResult()
        matches_nodeset.append(rtf.firstChild)
        match = compiled.search(arg, match.end())
    return matches_nodeset


def Map(context, funcname, *nodesets):
    """
    Apply the function serially over the given node sets.
    In iteration i, the function is passed N parameters
    where N is the number of argument node sets.  Each
    parameter is a node set of size 1, whose node is
    the ith node of the corresponding argument node set.
    The return value is a node set consisting of a series
    of result-tree nodes, each of which is a text node
    whose value is the string value of the result of the
    ith function invocation.
    Warning: this function uses the implied ordering of the node set
    Based on its implementation as a Python list.  But in reality
    There is no reliable ordering of XPath node sets.
    Therefore this function is not recommended for use with
    more than one node set parameter.
    """
    (prefix, local) = SplitQName(funcname)
    uri = context.processorNss.get(prefix)
    if prefix and not uri:
        raise XsltRuntimeException(Error.UNDEFINED_PREFIX,
                                   context.currentInstruction, prefix)
    expanded = (prefix and uri or '', local)
    func = context.functions.get(expanded)
    if not func:
        raise Exception('Dynamically invoked function %s not found.'%funcname)

    flist = [func]*len(nodesets)
    lf = lambda x, f, *args: apply(f, args)
    retlist = apply(map, (lf, flist) + nodesets)

    processor = context.processor
    result_nodeset = []
    for ret in retlist:
        processor.pushResult()
        processor.writers[-1].text(Conversions.StringValue(ret))
        rtf = processor.popResult()
        context.rtfs.append(rtf)
        result_nodeset.append(rtf.firstChild)
    return result_nodeset


def Lookup(context, name, key):
    """
    f:lookup() queries an index as defined by f:create-index.
    """
    name = Conversions.StringValue(name)
    key = Conversions.StringValue(key)
    processor = context.processor
    indices = processor.extensionParams.get((FT_EXT_NAMESPACE, 'indices'), {})
    index = indices.get(name, {})
    value = index.get(key, [])
    return value


ExtNamespaces = {
    FT_EXT_NAMESPACE : 'f',
    }

ExtFunctions = {
    (FT_EXT_NAMESPACE, 'key') : FtKey,
    (FT_EXT_NAMESPACE, 'lookup') : Lookup,
    (FT_EXT_NAMESPACE, 'search-re') : SearchRe,
    (FT_EXT_NAMESPACE, 'map') : Map,
    }

