########################################################################
# $Source: /var/local/cvsroot/4Suite/Ft/Xml/Xslt/Exslt/Strings.py,v $ $Revision: 1.6 $ $Date: 2002/09/06 03:43:37 $
"""
EXSLT - Strings

Copyright 2002 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

import urllib
from Ft.Xml.XPath import Conversions
from Ft.Xml.Xslt import XsltRuntimeException, Error
from Ft.Xml.Xslt.CopyOfElement import CopyNode

EXSL_STRINGS_NS = "http://exslt.org/strings"

def Align(context, target, padding, alignment=''):
    """
    The str:align function aligns a string within another string.

    See http://exslt.org/str/functions/align/str.align.html for further
    explanation.
    """
    target = Conversions.StringValue(target)
    padding = Conversions.StringValue(padding)
    alignment = alignment and Conversions.StringValue(alignment)

    # If the target string is longer than the padding string, then it is
    # truncated to be the same length as the padding string and returned.
    if len(target) > len(padding):
        return target[:len(padding)]

    # If no third argument is given or if it is not one of 'left', 'right'
    # or 'center', then it defaults to left alignment.
    if alignment == 'right':
        result = padding[:-len(target)] + target
    elif alignment == 'center':
        # With center alignment, the range of characters replaced by the target
        # string is in the middle of the padding string, such that either the
        # number of unreplaced characters on either side of the range is the
        # same or there is one less on the left than there is on the right.
        left = (len(padding) - len(target)) / 2
        right = left + len(target)
        result = padding[:left] + target + padding[right:]
    else:
        result = target + padding[len(target):]
    return result

def Concat(context, nodeset):
    """
    The str:concat function takes a node set and returns the concatenation of
    the string values of the nodes in that node set. If the node set is empty,
    it returns an empty string.
    """
    if type(nodeset) != type([]):
        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE, 
                                   context.currentInstruction)

    strings = map(Conversions.StringValue, nodeset)
    return u''.join(strings)

def DecodeUri(context, uri):
    """
    The str:decode-uri function decodes a string that has been URI-encoded.

    The function does not encode the following characters:
    ":", "/", ";", and "?". 
    """
    return urllib.unquote(Conversions.StringValue(uri))

def EncodeUri(context, uri):
    """
    The str:encode-uri function returns an encoded URI.
    """
    return urllib.quote(Conversions.StringValue(uri))

def Padding(context, length, chars=''):
    """
    The str:padding function creates a padding string of a certain length.

    The second argument gives a string to be used to create the padding.
    This string is repeated as many times as is necessary to create a string
    of the length specified by the first argument; if the string is more than
    a character long, it may have to be truncated to produce the required
    length. If no second argument is specified, it defaults to a space (' ').
    """
    length = Conversions.NumberValue(length)
    chars = chars and Conversions.StringValue(chars) or u' '

    return (chars*length)[:length]

def Replace(context, string, search, replace):
    """
    The str:replace function replaces any occurrences of search strings
    within a string with replacement nodes to create a node set.

    See http://exslt.org/str/functions/replace/str.replace.html for details.
    """
    if type(search) is not type([]):
        search_set = [Conversions.StringValue(search)]
    else:
        search_set = map(Conversions.StringValue, search)

    if type(replace) is not type([]):
        replace = [Conversions.StringValue(replace)]

    replacements = map(None, search_set, replace)
    replacements.sort(lambda a, b: cmp(len(a[0]), len(b[0])))

    processor = context.processor
    processor.pushResult()
    for (search, replace) in replacements:
        position = search.find(replace)
        if position >= 0:
            if position > 0:
                processor.writers[-1].text(string[:position])
            replace and CopyNode(processor, replace)
            string = string[position+len(search):]
    rtf = processor.popResult()
    return rtf.childNodes

def Split(context, string, pattern=u' '):
    """
    The str:split function splits up a string and returns a node set of
    token elements, each containing one token from the string.

    The first argument is the string to be split. The second argument is a
    pattern string (default=' '). The string given by the first argument is
    split at any occurrence of this pattern. An empty string pattern will
    result in a split on every character in the string.
    """
    string = Conversions.StringValue(string)
    pattern = Conversions.StringValue(pattern)
    processor = context.processor
    processor.pushResult()
    if pattern:
        for token in string.split(pattern):
            processor.writers[-1].startElement(u'token')
            processor.writers[-1].text(token)
            processor.writers[-1].endElement(u'token')
    else:
        for ch in string:
            processor.writers[-1].startElement(u'token')
            processor.writers[-1].text(ch)
            processor.writers[-1].endElement(u'token')
    rtf = processor.popResult()
    return rtf.childNodes

def Tokenize(context, string, delimiters=''):
    """
    The str:tokenize function splits up a string and returns a node set of
    'token' elements, each containing one token from the string.

    The first argument is the string to be tokenized. The second argument
    is a string consisting of a number of characters. Each character in
    this string is taken as a delimiting character. The string given by the
    first argument is split at any occurrence of any of these characters.
    """
    string = Conversions.StringValue(string)
    delim = delimiters and Conversions.StringValue(delimiters) or '\t\r\n '

    processor = context.processor
    processor.pushResult()
    token = u''
    for ch in string:
        if ch not in delim:
            token += ch
        elif token:
            # write out characters gathered up to now
            processor.writers[-1].startElement(u'token')
            processor.writers[-1].text(token)
            processor.writers[-1].endElement(u'token')
            token = u''
    if token:
        processor.writers[-1].startElement(u'token')
        processor.writers[-1].text(token)
        processor.writers[-1].endElement(u'token')
    rtf = processor.popResult()
    return rtf.childNodes

ExtNamespaces = {
    EXSL_STRINGS_NS : 'str',
    }

ExtFunctions = {
    (EXSL_STRINGS_NS, 'align') : Align,
    (EXSL_STRINGS_NS, 'concat') : Concat,
    (EXSL_STRINGS_NS, 'decode-uri') : DecodeUri,
    (EXSL_STRINGS_NS, 'encode-uri') : EncodeUri,
    (EXSL_STRINGS_NS, 'padding') : Padding,
    (EXSL_STRINGS_NS, 'replace') : Replace,
    (EXSL_STRINGS_NS, 'split') : Split,
    (EXSL_STRINGS_NS, 'tokenize') : Tokenize,
}

ExtElements = {}
