########################################################################
#
# File Name:            Processor.py
#
# Documentation:        http://docs.4suite.org/4XSLT/Processor.py.html
#
"""
Implement the XSLT processor engine
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os, sys, types, warnings, operator
#import traceback
from xml.dom import Node
from Ft import STRING_TYPES, DEFAULT_ENCODING
from Ft.Xml import InputSource, EMPTY_NAMESPACE
from Ft.Xml.Lib import StripElements
from Ft.Xml.XPath import Util
from Ft.Xml.Xslt import XSL_NAMESPACE, XsltContext
from Ft.Xml.Xslt import RtfWriter, OutputHandler, Error, XsltException
from Ft.Xml.Xslt import StylesheetReader
from Ft.Xml import Domlette
#from Ft.Lib import FtWarnings
XSLT_IMT = ['text/xml', 'application/xml']

# For builtin extension elements/functions
import Exslt, BuiltInExtElements

class Processor:
    _4xslt_trace = 0
    _4xslt_profile = 0
    _4xslt_debug = 0

    def __init__(self,
                 stylesheetIncPaths=None,
                 useEnvironment=1,
                 documentReader = None
                 ):
        self._suppressMsgs = 0
        self.stylesheet = None
        self._stylesheetIncPaths = stylesheetIncPaths or []
        if useEnvironment:
            if os.environ.has_key('XSLTINCLUDE'):
                self._stylesheetIncPaths.extend(
                    os.environ["XSLTINCLUDE"].split(os.pathsep)
                    )

        #Create the default reader for documents
        self._docReader = documentReader or Domlette.NonvalidatingReader

        self._lastOutputParams = None

        # has the "built-in template invoked with params" warning been issued?
        self._builtInWarningGiven = 0

        self.extFunctions = {}  #Cache ext functions to give to the context

        self.extElements = {}
        self.extElements.update(Exslt.ExtElements)
        self.extElements.update(BuiltInExtElements.ExtElements)

        # Create a single instance of this to be used every where
        self._styReader = StylesheetReader.StylesheetReader(
            self._stylesheetIncPaths
            )
        return

    def getStripElements(self):
        if self.stylesheet:
            return self.stylesheet.spaceRules
        else:
            return []

    def registerExtensionModules(self, moduleList):
        """
        Register a list of modules that support the public ExtFunction and/or
        ExtElement dictionaries
        """
        for mod_name in moduleList:
            if mod_name:
                mod = __import__(mod_name,{},{},['ExtFunctions'])
                if hasattr(mod,'ExtFunctions'):
                    self.extFunctions.update(mod.ExtFunctions)
                if hasattr(mod,'ExtElements'):
                    self.extElements.update(mod.ExtElements)


    def registerExtensionFunction(self, namespace, localName, function):
        """
        Register a single extension function
        """
        self.extFunctions[(namespace, localName)] = function

    def registerExtensionElement(self, namespace, localName, klass):
        """
        Register a single extension element
        """
        self.extElements[(namespace, localName)] = klass

    def setDocumentReader(self, docReader):
        """
        Set the reader used to read in an input source.
        Commonly used to set the reader to Domlette.ValidatingReader, but anything with a
        parse interface that returns a DOM tree will work
        """
        self._docReader = docReader

    def appendStylesheet(self, iSrc):
        """
        Append a stylesheet from an input source
        """
        sty = self._styReader.fromSrc(iSrc,extElements = self.extElements)
        if self.stylesheet:
            for child in self.stylesheet.children:
                sty.appendChild(child)
            sty.setup()
        self.stylesheet = sty
        self.outputParams = self.stylesheet.outputParams
        return

    def appendStylesheetInstance(self, sty, refUri=''):
        '''
        Accepts a valid StyleDOM node.  We need a refUri because instant stylesheets
        can be moved around, therefore changing how imports and includes work
        '''
        if self.stylesheet:
            sty = self._styReader.fromInstant(sty, refUri, is_import=1)
            for child in self.stylesheet.children:
                sty.appendChild(child)
            sty.setup()
        self.stylesheet = sty
        self.outputParams = self.stylesheet.outputParams
        return

    def run(self, iSrc, ignorePis=0, topLevelParams=None,
            writer=None, outputStream=None):
        #Update the strip elements
        #Assume that the ones from XSLT have higher priority
        ns = self.getStripElements()
        for s in iSrc.stripElements:
            ns.append(s)
        iSrc.stripElements = ns
        self._documentInputSource = iSrc
        try:
            src = self._docReader.parse(iSrc)
        except Exception, e:
            raise XsltException(Error.SOURCE_PARSE_ERROR, iSrc.uri or '<Python string>', e)
        if not ignorePis and self.checkStylesheetPis(src, iSrc):
            #Do it again with updates WS strip lists

            #NOTE:  There is a case where this will produce the wrong results.  If, there were
            #previous stylesheets that defined removing white space, then the
            #processing instruction referenced a stylesheet that overrode some of these
            #whitespace processing rules, the original trimmed space will be lost

            #Regardless, we need to remove any new whitespace defined in the PI
            self._stripElements(src)

        return self.execute(src, iSrc, ignorePis, topLevelParams,
                            writer, outputStream)

    def runNode(self, node, sourceUri, ignorePis=0, topLevelParams=None,
                writer=None, outputStream=None,preserveSrc=0):
        """
        Execute a source document from a DOM:

        node is the node to be executed
        sourceUri is the URI of the source DOM tree.
        ignorePis signals that we should ignore Processing Instructions in the stylesheet(s)
        topLevelParams is a dictionary of top level parameter definitions keyed like
          {(namespace,localname)} = value
        writer is an instance of an output writer
        outputStream allows redirection of the output
        preserverSrc signals that the source DOM should not be mutated
        """

        if node.nodeType != Domlette.Node.DOCUMENT_NODE:
            raise ValueError("Node must be a document")

        if preserveSrc:
            newDoc = Domlette.implementation.createDocument(None,None,None)
            for child in node.childNodes:
                new_node = newDoc.importNode(child,1)
                newDoc.appendChild(new_node)
                node = newDoc

        self._stripElements(node)

        #Create a dummy iSrc
        docInputSource = InputSource.InputSource(None,
                                                 sourceUri,
                                                 processIncludes=1,
                                                 stripElements=self.getStripElements(),
                                                 factory=InputSource.DefaultFactory)


        if not ignorePis and self.checkStylesheetPis(node, docInputSource):
            #Do it again with updates WS strip lists

            #NOTE:  There is a case where this will produce the wrong results.  If, there were
            #previous stylesheets that defined removing white space, then the
            #processing instruction referenced a stylesheet that overrode some of these
            #whitespace processing rules, the original trimmed space will be lost
            
            #Regardless, we need to remove any new whitespace defined in the PI
            self._stripElements(node)


        return self.execute(node,
                            docInputSource,
                            ignorePis=ignorePis,
                            topLevelParams=topLevelParams,
                            writer=writer,
                            outputStream=outputStream)

    def checkStylesheetPis(self, node, inputSource):
        #Note: A Stylesheet PI can only be in the prolog, acc to the NOTE
        #http://www.w3.org/TR/xml-stylesheet/

        root = node.ownerDocument or node

        hrefs = []
        for child in root.childNodes:
            if child.nodeType == Node.PROCESSING_INSTRUCTION_NODE:
                if child.target == 'xml-stylesheet':
                    data = child.data.split()
                    pseudo_attrs = {}
                    for d in data:
                        seg = d.split('=')
                        if len(seg) == 2:
                            pseudo_attrs[seg[0]] = seg[1][1:-1]

                    if pseudo_attrs.has_key('href'):
                        imt = pseudo_attrs.get('type')
                        if imt is None or imt in XSLT_IMT:
                            hrefs.append(pseudo_attrs['href'])

        if hrefs:
            self._styReader._import_index += 1
            for href in hrefs:
                # Resolve the PI from the document input Source
                new_source = inputSource.resolve(href, '', 'xml-stylesheet PI')
                self.appendStylesheet(new_source)

        # Return true if xml-stylesheet processing instructions are handled
        return not not hrefs

    def execute(self, node, docInputSource, ignorePis=0, topLevelParams=None,
                writer=None, outputStream=None):
        '''
        Run the stylesheet processor against the given XML DOM node with the
        stylesheets that have been registered.  Does not mutate the DOM
        If writer is None, use the XmlWriter, otherwise, use the supplied
        writer.
        '''
        #FIXME: What about ws stripping?
        topLevelParams = topLevelParams or {}

        self.attributeSets = {}
        self.keys = {}

        if not self.stylesheet:
            raise XsltException(Error.NO_STYLESHEET)

        if not writer:
            # Use OutputHandler to determine the real writer to use.
            writer = OutputHandler.OutputHandler(self.outputParams,
                                                 outputStream,
                                                 self.__writerChanged)
        self.writers = [writer]

        # Setup the named templates
        self._namedTemplates = self.stylesheet.getNamedTemplates()

        # Initialize any stylesheet parameters
        tlp = topLevelParams.copy()
        self._normalizeParams(tlp)
        self._docInputSource = docInputSource

        # Prepare the stylesheet for processing
        self.stylesheet.prime(node, self, tlp, docInputSource.uri)
        self.stylesheet.updateKeys(node, self)

        # Create the context used for processing
        variables = self.stylesheet.getGlobalVariables()
        functions = self.stylesheet.getInitialFunctions()
        context = XsltContext.XsltContext(node, 1, 1, None,
                                          varBindings=variables,
                                          processor=self,
                                          extFunctionMap=functions)
        context.addDocument(node, docInputSource.uri)

        # Process the document
        self.writers[-1].startDocument()
        self.applyTemplates(context)
        self.writers[-1].endDocument()
        result = self.writers[-1].getResult()

        # Perform cleanup
        self.stylesheet.idle(node, self, docInputSource.uri)
        Util.FreeDocumentIndex(node)

        #How does this contrast with access to self.outputParams ?
        self._lastOutputParams = self.writers[-1]._outputParams

        return result

    def applyTemplates(self, context, params=None):
        params = params or {}
        if not self.stylesheet.applyTemplates(context, self, params):
            # No matching templates found, use builtin templates
            if params and not self._builtInWarningGiven:
                self.warning('Built-in template invoked with params that '
                             'will be ignored. This message will only '
                             'appear once per transform.')
                self._builtInWarningGiven = 1
            if context.node.nodeType == Node.TEXT_NODE:
                self.writers[-1].text(context.node.data)
            elif context.node.nodeType in [Node.ELEMENT_NODE, Node.DOCUMENT_NODE]:
                state = context.copy()
                node_set = context.node.childNodes
                size = len(node_set)
                pos = 1
                for node in node_set:
                    context.node, context.position, context.size = \
                                  node, pos, size
                    self.applyTemplates(context)
                    pos += 1
                context.set(state)
            elif context.node.nodeType == Node.ATTRIBUTE_NODE:
                self.writers[-1].text(context.node.value)
        return

    # FIXME: l10n
    def xslMessage(self, msg):
        if not self._suppressMsgs:
            sys.stderr.write("STYLESHEET MESSAGE:\n")
            sys.stderr.write(msg+'\n')
            sys.stderr.write("END STYLESHEET MESSAGE\n")
        return

    # FIXME: l10n
    def warning(self, message):
        if not self._suppressMsgs:
            warnings.warn(message)
            #FtWarnings.warn(message, FtWarnings.RuntimeWarning)
        return

    def messageControl(self, suppress):
        self._suppressMsgs = suppress
        return

    def __writerChanged(self, newWriter):
        """
        Private callback function to allow the OutputHandler to remove
        itself from the writer stack.
        """
        self.writers[-1] = newWriter
        return

    def addHandler(self, outputParams, stream):
        """
        Add the handler used to determine real writer used for a particular
        output method.
        """
        handler = OutputHandler.OutputHandler(outputParams, stream,
                                              self.__writerChanged)
        self.writers.append(handler)
        handler.startDocument()
        return

    def removeHandler(self):
        self.writers[-1].endDocument()
        del self.writers[-1]

    def pushResult(self, handler=None, ownerDoc=None):
        '''
        Start processing all content into a separate result-tree
        (either an rtf, or for ft:write-file/common:document)
        '''
        # FIXME: Should actually use a doc fragment for the SAX handler doc
        # Q: Should the output parameters discovered at run-time
        #   (e.g html root element) be propagated back to RTFs?
        handler = handler or RtfWriter.RtfWriter(self.outputParams)
        self.writers.append(handler)
        return

    def popResult(self):
        '''End sub-result-tree and return any result'''
        result = self.writers[-1].getResult()
        del self.writers[-1]
        return result

    def output(self):
        """Returns the current output writer"""
        return self.writers[-1]

    def releaseRtf(self, rtf):
        RtfWriter.ReleaseRtf(rtf)
        return

    def _stripElements(self,node):
        StripElements.StripElements(node, self.getStripElements())
        return

    def _normalizeParams(self, params):
        """
        params is a dictionary of top-level parameters.  The main task is to
        check this dictionary for lists of strings and convert these to
        a node set of text nodes
        """
        for k, v in params.items():
            if v and isinstance(v, types.StringType):
                try:
                    # Try UTF-8
                    params[k] = unicode(v, 'UTF-8')
                except ValueError:
                    # Use encoding from locale
                    try:
                        params[k] = unicode(v, DEFAULT_ENCODING)
                    except ValueError:
                        raise Exception(
                            "String parameters must be Unicode objects or "
                            "strings encoded as UTF-8 or %s." %
                            DEFAULT_ENCODING)
            elif v and type(v) == type([]) and type(v[0]) in STRING_TYPES:
                handler = RtfWriter.RtfWriter()
                for text in v:
                    if isinstance(text, types.StringType):
                        # Convert it to unicode
                        try:
                            # Try UTF-8
                            text = unicode(text, 'UTF-8')
                        except ValueError:
                            # Use encoding from locale
                            try:
                                text = unicode(text, DEFAULT_ENCODING)
                            except ValueError:
                                raise Exception(
                                    "String parameters must be Unicode objects "
                                    "or strings encoded as UTF-8 or %s." %
                                    DEFAULT_ENCODING)
                    handler.text(text)
                rtf = handler.getResult()
                new_v = rtf.childNodes
                del params[k]
                params[k] = new_v
        return

    def reclaim(self):
        return

    def reset(self):
        self.stylesheet = None
        self._styReader = StylesheetReader.StylesheetReader(self._stylesheetIncPaths)
        return
