########################################################################
#
# File Name:           StylesheetReader.py
#
"""
Create a stylesheet object
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 1999-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""


from Ft.Xml import VerifyPyexpat
VerifyPyexpat("StylesheetReader.py")

import urllib, cStringIO, re
#import codecs
from xml.dom import Node
from xml.parsers import expat

from Ft.Lib import Uri
from Ft.Xml import Xslt, XMLNS_NAMESPACE, InputSource, VerifyPyexpat
from Ft.Xml.Xslt import XsltException, Error

from StylesheetHandler import NS_SPLIT_CHAR, StylesheetHandler

# borrowed from xmllib (there are too many regexp to import it)
_S = '[ \t\r\n]+'                       # white space
_opS = '[ \t\r\n]*'                     # optional white space
_QStr = "(?:'[^']*'|\"[^\"]*\")"        # quoted XML string
_xmldecl = re.compile('<\?xml' + _S + 'version' + _opS + '=' + _opS + _QStr +
                      '(?:' + _S + 'encoding' + _opS + '=' + _opS +
                      "(?P<encoding>'[A-Za-z][-A-Za-z0-9._]*'|"
                      '"[A-Za-z][-A-Za-z0-9._]*"))?'
                      '(?:' + _S + 'standalone' + _opS + '=' + _opS +
                      '(?:\'(?:yes|no)\'|"(?:yes|no)"))?' + _opS + '\?>')

def _change_node(xslt_node, attributes, importIndex=0):
    xslt_node.__dict__.update(attributes)
    xslt_node.importIndex += importIndex
    for child in xslt_node.children or []:
        _change_node(child, attributes, importIndex)
    return

class StylesheetReader(StylesheetHandler):

    def __init__(self, includePaths=None, ownerDocument=None,
                 importIndex=0, globalVars=None, extElements=None):
        StylesheetHandler.__init__(self, importIndex, globalVars, extElements)
        self._include_paths = includePaths or []
        self._ownerDoc = ownerDocument
        return

    def reset(self):
        StylesheetHandler.reset(self)
        self._ownerDoc = None
        self._input_source = None
        return

    def clone(self):
        # This method is provided to allow for ease of subclassing
        return self.__class__(self._include_paths, self._ownerDoc,
                              self._import_index, self._global_vars,self._extElements)

    def createParser(self, baseUri):
        parser = expat.ParserCreate(namespace_separator=NS_SPLIT_CHAR)
        parser.StartElementHandler = self.startElement
        parser.EndElementHandler = self.endElement
        parser.StartNamespaceDeclHandler = self.startNamespaceDecl
        parser.EndNamespaceDeclHandler = self.endNamespaceDecl
        parser.CharacterDataHandler = self.characterData
        parser.ExternalEntityRefHandler = self.externalEntityRef
        parser.SetBase(baseUri)
        return parser

    def fromInstant(self, instant, baseUri='',is_import=0):
        root = instant.root
        if baseUri and root.baseUri != baseUri:
            update_attrs = {'baseUri' : baseUri}
        else:
            update_attrs = {}

        if self._ownerDoc:
            # this is an inclusion; change the document and importIndex
            update_attrs['ownerDocument'] = self._ownerDoc
            _change_node(root, update_attrs, self._import_index + (is_import and 1 or 0))
            self._ownerDoc.primeInstructions.extend(root.primeInstructions)
            self._ownerDoc.idleInstructions.extend(root.idleInstructions)
        elif update_attrs:
            _change_node(root, update_attrs)

        # Update the state of the reader
        if not self._ownerDoc:
            self._ownerDoc = root
        self._import_index = root.stylesheet.importIndex

        return root.stylesheet

    def fromDocument(self, document, baseUri='',isf=None):

        if not baseUri:
            if hasattr(document, 'baseUri'):
                # DOM Level 3
                baseUri = document.baseUri
            elif hasattr(document, 'refUri'):
                # cDomlette
                baseUri = document.refUri

        self._baseUri = baseUri

        #We need to fake out an input source on the document
        isf = isf or InputSource.DefaultFactory
        self._input_source = isf.fromStream(None,baseUri)

        # This is needed to handle XIncludes
        expat_parser = self.createParser(baseUri)
        self.parser = DomParser(expat_parser)
        self.startDocument()
        try:
            self.parser.parse(document)
        except expat.ExpatError, error:
            raise XsltException(Error.STYLESHEET_PARSE_ERROR, self._baseUri,
                                error.lineno, error.offset,
                                expat.ErrorString(error.code))
        self.endDocument()
        root = self._node_stack[0]
        root.stylesheet._input_source = self._input_source

        if root is self._ownerDoc:
            # the top-most stylesheet
            root.stylesheet.setup()

        return root.stylesheet

    def fromSrc(self, src,extElements = None):
        """
        Read in a stylesheet from an InputSource
        """
        #FIXME: This is very wasteful in the case that document("") is not called,
        # because we need to cache the original content of the stylesheet for document("")
        # when we read in the stream, then create a new InputSource with the content

        content = src.stream.read()
        new_stream = cStringIO.StringIO(content)
        new_source = src._clone(new_stream,src.uri)

        #Set up the ext elements if given
        if extElements is not None:
            self._extElements.update(extElements)

        #Do the parsing
        stylesheet =  self._parseSrc(new_source)
        
        #Do the caching
        self._ownerDoc.sources[src.uri] = content

        return stylesheet

    def _parseSrc(self, src):

        self._input_source = src

        self.parser = self.createParser(src.uri)
        self.startDocument()
        try:
            success = self.parser.ParseFile(src.stream)
        except expat.ExpatError, error:
            raise XsltException(Error.STYLESHEET_PARSE_ERROR, src.uri,
                                error.lineno, error.offset,
                                expat.ErrorString(error.code))
        self.endDocument()
        root = self._node_stack[0]
        root.stylesheet._input_source = self._input_source

        if root is self._ownerDoc:
            # the top-most stylesheet
            root.stylesheet.setup()

        return root.stylesheet


class DomParser:
    """
    A class which converts a DOM tree into the corresponding expat parser
    callbacks. This allows for all stylesheet logic to exist in one class.
    """

    def __init__(self, parser):
        self._parser = parser

        # mimic expat parser object interface
        self.ErrorLineNumber = '??'
        self.ErrorColumnNumber = '??'
        return

    # mimic expat parser object interfaces
    def GetBase(self):
        return self._parser.GetBase()

    # mimic expat parser object interfaces
    def SetBase(self, baseUri):
        return self._parser.SetBase(baseUri)

    # mimic expat parser object interfaces
    def ExternalEntityParserCreate(self, context):
        return self._parser.ExternalEntityParserCreate(context)

    def parse(self, document):
        if document.documentElement:
            self._current_nss = {}
            self._walk_tree(document.documentElement)
        return

    def _walk_tree(self, node):
        if node.nodeType == Node.ELEMENT_NODE:
            namespaces = {}
            attrs = {}
            # Create expat-style attribute names and trim out namespaces
            for attr in node.attributes.values():
                namespace = attr.namespaceURI
                if namespace == XMLNS_NAMESPACE:
                    # actual namespace declarations
                    if attr.prefix:
                        # xmlns:prefix
                        prefix = attr.localName
                    else:
                        prefix = None

                    namespaces[prefix] = attr.value or None
                else:
                    if namespace:
                        # DOM doesn't need separate namespace declarations
                        namespaces[attr.prefix] = namespace
                        name = namespace + NS_SPLIT_CHAR + attr.localName
                    else:
                        name = attr.localName
                    attrs[name] = attr.value

            # Create expat-style element name
            if node.namespaceURI:
                # DOM doesn't need separate namespace declarations
                namespaces[node.prefix] = node.namespaceURI
                name = node.namespaceURI + NS_SPLIT_CHAR + node.localName
            else:
                name = node.localName

            # StartNamespaceDeclHandler
            current = self._current_nss.copy()
            for prefix, namespace in namespaces.items():
                if current.get(prefix, -1) == namespace:
                    # remove duplicate definitions
                    del namespaces[prefix]
                else:
                    self._current_nss[prefix] = namespace
                    self._parser.StartNamespaceDeclHandler(prefix, namespace)

            # ElementHandler
            self._parser.StartElementHandler(name, attrs)
            for child in node.childNodes:
                self._walk_tree(child)
            self._parser.EndElementHandler(name)

            # EndNamespaceDeclHandler
            for prefix in namespaces.keys():
                self._parser.EndNamespaceDeclHandler(prefix)

            # restore original namespaces
            self._current_nss = current

        elif node.nodeType in [Node.TEXT_NODE, Node.CDATA_SECTION_NODE]:
            # CDataSection is just another way of specifying text
            self._parser.CharacterDataHandler(node.data)

        # XSLT ignores everything except Text and Element nodes
        return

    
